//**********************************************************************
//
//              AUDIOSG.c          I[fBIM@2007/Mar/11
//
//**********************************************************************

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <avr/eeprom.h>
#include "sg.h"
#include "sin_tbl.h"

#define sbi(sfr, bit)    (_SFR_BYTE(sfr) |= _BV(bit))
#define cbi(sfr, bit)    (_SFR_BYTE(sfr) &= ~_BV(bit))

#define true  1
#define false 0

#define NOP asm volatile("nop\t\n")

#define LF       (0x0a)       // \n
#define CR       (0x0d)       // \r

// -------- n[hEFAݒ --------------------------------------------------
#define FCPU      24576000

#define DA_PORT   PORTD
#define DA_LRCLK  2

#define SW_PORT   PIND
#define SW_LEFT   3
#define SW_DOWN   5
#define SW_UP     6
#define SW_RIGHT  7
#define SW_MASK   0xE8

#define LCDC_DATA PORTC
#define LCDC_CTRL PORTC
#define LCDC_E    5
#define LCDC_RS   4

#define PG_PORT   PORTB
#define PG_OUT    1

#define TEST_PORT PORTB
#define TEST_OUT  2

// -------- c`Ro[^[Eof֘A̐ݒ ------------------------------------
#define FS       (FCPU/128)
#define LRCLK_H  sbi(DA_PORT,DA_LRCLK)
#define LRCLK_L  cbi(DA_PORT,DA_LRCLK)

#define PG_H     sbi(PG_PORT,PG_OUT)
#define PG_L     cbi(PG_PORT,PG_OUT)

// -------- ͂rv֘A̐ݒ ------------------------------------------------
#define IS_LEFT_PRESSED    bit_is_clear(SW_PORT,SW_LEFT)
#define IS_DOWN_PRESSED    bit_is_clear(SW_PORT,SW_DOWN)
#define IS_UP_PRESSED      bit_is_clear(SW_PORT,SW_UP)
#define IS_RIGHT_PRESSED   bit_is_clear(SW_PORT,SW_RIGHT)
#define IS_ANYSW_PRESSED   ((~(uint8)SW_PORT))&((uint8)0xE8)

// -------- kbc֘A̐ݒ ----------------------------------------------
#define LCDC_MAX_X      (16)            // kbcW[  
#define LCDC_MAX_Y      (2)             // kbcW[ s@őSs

#define LCDC_NOCURSOR   (0x0C)          // J[\@Ȃ
#define LCDC_BLINK      (0x0D)          // J[\@uN[h
#define LCDC_UNDERLINE  (0x0E)          // J[\@[h

#define LCDC_E_HIGH()   sbi( LCDC_CTRL, LCDC_E )
#define LCDC_E_LOW()    cbi( LCDC_CTRL, LCDC_E )
#define LCDC_RS_HIGH()  sbi( LCDC_CTRL, LCDC_RS )
#define LCDC_RS_LOW()   cbi( LCDC_CTRL, LCDC_RS )

static uint8 x, y;

// -------- ̑ݒ ----------------------------------------------------
#define TEST_H  sbi(TEST_PORT,TEST_OUT)
#define TEST_L  cbi(TEST_PORT,TEST_OUT)

// -------- O[oϐ ------------------------------------------
PARAM pa;

int8   step=1,
       mode=0,                // 0:wave 1:memrd 2:memwr 3:freqstep
       mem=1,                 // [ԍ
       freqstep=4,            // 0:1Hz 1:10Hz 2:100Hz 3:1-1.1 4:1-2-3 5:1-2-5 6:1-10
       phasestep=5,           // 0:1 1:5 2:10 3:30 4:45 5:90
       levelstep=3;           // 0:1 1:2 2:5  3:10 4:20

PARAM  pa_eeprom[31] __attribute__((section(".eeprom")));

char   str[34];               // kbc\p

uint8 phasesteptbl[6] = { 1, 5, 10, 30, 45, 90 };
uint8 levelsteptbl[5] = { 1, 2, 5, 10, 20 };

prog_int16_t *psin2048_tbl = (prog_int16_t *)sin2048_tbl;

volatile PARAMSG pasg;        // SG ̐ݒl
PARAMSG *paramsg = &pasg;

// -------- vOǂݍ ------------------------------------------
#include "util.c"
#include "lcdc16x2.c"


// ------------------------------------------------------------------------
// n[hEFA
// ------------------------------------------------------------------------
void avr_init(void) {
  DDRB = 0x17;                // 5:SCK 4:MISO 3:MOSI 2:TEST 1:PGOUT 0:OSC_OUT
  DDRC = 0x3F;                // 5:E 4:RS 3-0:DB7-4
  DDRD = 0x16;                // 7:SW4 6:SW3 5:SW2 4:BCK 3:SW1 2:LRCLK 1:DATA 0:
  
  PORTD = 0xE8;               // pullup for sw0-3
  
  UBRR0 = 0;                  // 0:fclk/2 1:fclk/4
  UCSR0C = (1<<UMSEL01) | (1<<UMSEL00) | (1<<UCPHA0) | (1<<UCPOL0);
  UCSR0B = (1<<TXEN0);
  
  return;
}


// ------------------------------------------------------------------------
// g㉺@㉺Xebv freqstep ɂ
// ------------------------------------------------------------------------
void frequpdown( uint8 fn ) {
  uint16 base = 1;
  uint32 n = ( fn == 1 )? pa.freq1: pa.freq2;
  uint32 nmax;
  if( freqstep <= 2 ) {                 // 0:1Hz 1:10Hz 2:100Hz
    nmax = FS/2-1;
  }else if( freqstep == 3 ) {           // 3:1-1.1
    nmax = 99;
  }else{                                // 4:1-2-3 5:1-2-5 6:1-10
    nmax = 9;
  }
  while( n > nmax ) {
    n /= 10;
    base *= 10;
  }
  if( IS_UP_PRESSED ) {                 // UP
    if( freqstep == 0 ) {               // 0:1Hz
      n++;
    }else if( freqstep == 1 ) {         // 1:10Hz
      n += 10;
    }else if( freqstep == 2 ) {         // 2:100Hz
      n += 100;
    }else if( freqstep <= 4 ) {         // 3:1-1.1  4:1-2-3
      n++;
    }else if( freqstep == 5 ) {         // 5:1-2-5
      if( n == 1 ) {
        n = 2;
      }else if( n < 5 ) {
        n = 5;
      }else{
        n = 10;
      }
    }else{                              // 6:1-10
      n=10;
    }
  }else{                                // DOWN
    if( (n==1)||((freqstep==3)&&(n==10)) ) {
      if( base >= 10 ) {
				n *= 10;
        base /= 10;
      }
    }
    if( freqstep == 0 ) {               // 0:1Hz
      n--;
    }else if( freqstep == 1 ) {         // 1:10Hz
      n -= 10;
    }else if( freqstep == 2 ) {         // 2:10Hz
      n -= 100;
    }else if( freqstep <= 4 ) {         // 3:1-1.1  4:1-2-3
      n--;
    }else if( freqstep == 5 ) {         // 5:1-2-5
      if( n > 5 ) {
        n = 5;
      }else if( n >2 ) {
        n = 2;
      }else{
        n = 1;
      }
    }else{                              // 6:1-10
      n=1;
    }
  }
  uint32 freq = n*base;
  if( freq <= 0     ) freq = 1;
  if( freq > FS/2-1 ) freq = FS/2-1;
  if( fn == 1 ) {
    pa.freq1 = freq;
  }else{
    pa.freq2 = freq;
  }
}


// ------------------------------------------------------------------------
// x㉺
// ------------------------------------------------------------------------
void levelupdown( void ) {
  if( IS_UP_PRESSED ) {                 // UP
    pa.level -= levelsteptbl[levelstep];   // lقx͑傫̂
  }else{                                // DOWN
    pa.level += levelsteptbl[levelstep];
  }
  if( pa.level < 0  ) pa.level = 0;
  if( pa.level > 96 ) pa.level = 96;
}


// ------------------------------------------------------------------------
// sânmTCN㉺
// ------------------------------------------------------------------------
void oncycleupdown( void ) {
  if( IS_UP_PRESSED ) {                 // UP
    pa.oncycle++;
  }else{                                // DOWN
    pa.oncycle--;
  }
  if( pa.oncycle < 1  ) pa.oncycle = 99;
  if( pa.oncycle > 99 ) pa.oncycle = 1;
}


// ------------------------------------------------------------------------
// sâneeTCN㉺
// ------------------------------------------------------------------------
void offcycleupdown( void ) {
  if( IS_UP_PRESSED ) {                 // UP
    pa.offcycle++;
  }else{                                // DOWN
    pa.offcycle--;
  }
  if( pa.offcycle < 1  ) pa.offcycle = 99;
  if( pa.offcycle > 99 ) pa.offcycle = 1;
}


// ------------------------------------------------------------------------
// ʑ㉺
// ------------------------------------------------------------------------
void phaseupdown( void ) {
  if( IS_UP_PRESSED ) {                 // UP
    pa.phase += phasesteptbl[phasestep];
  }else{                                // DOWN
    pa.phase -= phasesteptbl[phasestep];
  }
  if( pa.phase < 0   ) pa.phase += 360;
  if( pa.phase > 359 ) pa.phase -= 360;
}


// ------------------------------------------------------------------------
// [ԍ㉺
// ------------------------------------------------------------------------
void memupdown( void ) {
  if( IS_UP_PRESSED ) {                 // UP
    mem++;
  }else{                                // DOWN
    mem--;
  }
  if( mem < 0  ) mem = 29;
  if( mem > 29 ) mem = 0;
}


// ------------------------------------------------------------------------
// [ǂݍ
// ------------------------------------------------------------------------
void memread( int8 m ) {
  eeprom_busy_wait();
  eeprom_read_block ( &pa, &pa_eeprom[m], sizeof( PARAM ) );
}


// ------------------------------------------------------------------------
// [ւ̏
// ------------------------------------------------------------------------
void memwrite( int8 m ) {
  eeprom_busy_wait();
  eeprom_write_block ( &pa, &pa_eeprom[m], sizeof( PARAM ) );
}



// ------------------------------------------------------------------------
// ݒG[ȂH
// ------------------------------------------------------------------------
bool checksetting( void ) {
  if( (pa.wave     < 0)||(pa.wave     > 8) )   return 1;    // 
  if( (pa.level    < 0)||(pa.level    > 99) )  return 1;    // 
  if( (pa.oncycle  < 1)||(pa.oncycle  > 99) )  return 1;    // 
  if( (pa.offcycle < 1)||(pa.offcycle > 99) )  return 1;    // 
  if(  pa.freq1    > FS/2 )                    return 1;    // 
  if(  pa.freq2    > FS/2 )                    return 1;    // 
  if( (pa.phase    < 0)||(pa.phase    > 359) ) return 1;    // 
  return 0;
}


// ------------------------------------------------------------------------
// ftH[gݒ
// ------------------------------------------------------------------------
void defaultsetting( void ) {
  pa.wave     = 0;       // 0:sine 1:Square 2:SawTooth 3:Triangle 4:ToneBurst 5:Sine 2ch 6:TwoTone 7:WhiteNoise 8:PinkNoise
  pa.level    = 0;       // 0:0dB 1:-1dB 2:-2dB ... 99:-99dB
  pa.oncycle  = 2;       // toneburst ̔g`ITCN
  pa.offcycle = 2;       // toneburst ̔g`ItTCN
  pa.freq1    = 1000,    // gP
  pa.freq2    = 2000;    // gQ
  pa.phase    = 90;      // gQ̏ʑ@PʁFx
}


// ------------------------------------------------------------------------
// SG ̐ݒlvZAݒ肷
// ------------------------------------------------------------------------
uint16 leveltbl[20] = {                 // 0dB ` -19dB
    65535, 58408, 52056, 46395, 41350, 36853, 32845, 29273, 26090, 23253,
    20724, 18470, 16462, 14671, 13076, 11654, 10387,  9257,  8250,  7353 };

void sgsetting( void ) {
  // SG ł̃xlݒ肷
  pasg.Levelsg=leveltbl[ pa.level % 20 ];         // 0dB ` -19dB ̓e[u
  int8 l = pa.level-20;
  pasg.Level256 = 0;
  while( l>=0) {                                  // 20dB Ƃ 1/10 ɂ
    if( pasg.Levelsg < 2560 ) {
      pasg.Levelsg = (pasg.Levelsg*16/10)*16;     // ȂƂ͂QTUŃXP[O
      pasg.Level256 = 1;
    }else{
      pasg.Levelsg /= 10;                         // 1/10 ɂ
    }
    l -= 20;
  }
  // g@PXrbg
  pasg.Freq1sgh = (uint8)(((uint32)pa.freq1*1024/(FS/512))>>16);   // 8bit
  pasg.Freq1sg  = (uint16)((uint32)pa.freq1*1024/(FS/512));        // 16bit
  pasg.Freq2sgh = (uint8)(((uint32)pa.freq2*1024/(FS/512))>>16);   // 8bit
  int16 m = pa.freq2/pa.freq1;
  if( m*pa.freq1 == pa.freq2 ) {  // freq2  freq1 ̐{ freq1 ɓ
    pasg.Freq2sg = m*pasg.Freq1sg;
  }else{
    pasg.Freq2sg = (uint16)((uint32)pa.freq2*1024/(FS/512));      // 16bit
  }
  // freq2 ̈ʑ  2048=360degree
  pasg.Phasesg  = (int16)((int32)2048*pa.phase/360);  
  // TCQł͂PTvOԕ̕␳Kv
  pasg.Phasesg_hosei = pasg.Phasesg - (int16)((int32)2048*pa.freq2/FS); 
  // 
  pasg.Oncycle  = pa.oncycle;
  pasg.Offcycle = pa.offcycle;
}

// ------------------------------------------------------------------------
// ꂽ SW ɉ
// ------------------------------------------------------------------------
void keyinput( void ) {
  if( IS_LEFT_PRESSED ) {
    if( step >= 2 ) {
      step = 1;
    }else if( step == 1 ){
      step = 0;
    }
  }else if( IS_RIGHT_PRESSED ) {
    if( step <= 1 ) {
      step++;
      if( ( step==1 )&&( mode == 1 ) ) memread( mem );   // [ݒǂݏo
      if( step==2 ) {
        if( mode == 2 ) memwrite( mem );   // [ւ̏
        mode = 0;           // step 2 ȍ~ wave
      }
    }else if( step == 2 ){
      if( pa.wave <= 6 ) {  // 0:Sine 1:Square 2:SawTooth 3:Triangle 4:ToneBurst 5:Sine 2ch 6:TwoTone
        step++;
      }
    }else if( step == 3 ) {
      if( pa.wave <= 3 ) {  // 0:Sine 1:Square 2:SawTooth 3:Triangle
        step = 2;
      }else{                // 4:ToneBurst 5:Sine 2ch 6:TwoTone
        step++;
      }
    }else if( step == 4 ) {
      step++;
    }else{
      step = 2;
    }
  }else if( step == 0 ) {
    if( IS_UP_PRESSED ) {             // UP
      mode++;
    }else{                            // DOWN
      mode--;
    }
    if( mode < 0 ) mode = 5;
    if( mode > 5 ) mode = 0;
  }else if( step == 1 ) {
    if( mode == 0 ) {                 // wave
      if( IS_UP_PRESSED ) {           // UP
        pa.wave++;
      }else{                          // DOWN
        pa.wave--;
      }
      if( pa.wave < 0 ) pa.wave = 8;
      if( pa.wave > 8 ) pa.wave = 0;
    }else if( mode == 1 ) {           // mem rd
      memupdown();
      memread( mem );                 // [ݒǂݏo
    }else if( mode == 2 ) {           // mem wr
      memupdown();
    }else if( mode == 3 ) {           // freq step
      if( IS_UP_PRESSED ) {           // UP
        freqstep++;
      }else{                          // DOWN
        freqstep--;
      }
      if( freqstep < 0 ) freqstep = 6;
      if( freqstep > 6 ) freqstep = 0;
    }else if( mode == 4 ) {           // phase step
      if( IS_UP_PRESSED ) {           // UP
        phasestep++;
      }else{                          // DOWN
        phasestep--;
      }
      if( phasestep < 0 ) phasestep = 5;
      if( phasestep > 5 ) phasestep = 0;
    }else{                            // level step
      if( IS_UP_PRESSED ) {           // UP
        levelstep++;
      }else{                          // DOWN
        levelstep--;
      }
      if( levelstep < 0 ) levelstep = 4;
      if( levelstep > 4 ) levelstep = 0;
    }
  }else if( step == 2 ) {             // g`
    if( pa.wave <= 6 ) {
      frequpdown( 1 );
    }else{
      levelupdown();
    }
  }else if( step == 3 ) {
    if( pa.wave <= 4 ) {
      levelupdown();
    }else{
      frequpdown( 2 );
    }
  }else if( step == 4 ) {
    if( pa.wave == 4 ) {
      oncycleupdown();
    }else{
      levelupdown();
    }
  }else if( step == 5 ) {
    if( pa.wave == 4 ) {
      offcycleupdown();
    }else{
      phaseupdown();
    }
  }else{
    levelupdown();
  }
}


// ------------------------------------------------------------------------
// LCD Ɍݏ\
// ------------------------------------------------------------------------
char * putcursor( char * s, uint8 st ) {
  if( step == st ) {
    return strcopy( s, "\001" ); // J[\}[N
  }else{
    return strcopy( s, " " );
  }
}

char * putfreq1( char * s ) {
  s = uint32str( pa.freq1, s, 5 );
  return strcopy( s, "Hz" );
}

char * modestr[6]  = { "Wave ", "MemRd ", "MemWr ", "FreqStep ", "PhaseStep ", "LevelStep " };
char * wavestr[9]  = { "Sine", "Square", "SawTooth", "Triangle", "ToneBurst", "Sine 2ch", "TwoTone", "WhiteNoise", "PinkNoise" };
char * waveshortstr[9]  = { "Si", "Sq", "Sa", "Tr", "Tb", "S2", "2T", "Wn", "Pn" };
char * freqstepstr[7] = { "1Hz", "10Hz", "100Hz", "1\1761.1", "1\1762\1763", "1\1762\1765", "1\17610" };

void disp( void ) {
  char * s;
  
  lcdc_cls();
  str[0]=0;
  
  if( (step>=2) && ((pa.wave==4)||(pa.wave==5)||(pa.wave==6)) ) { // toneburst sin2ch twotone
    if( pa.wave==4 ) {  // Toneburst
      s = putcursor( str, 2 );
      s = uint32str( pa.freq1, s, 5 ); s = strcopy( s, "Hz  " );
      s = putcursor( s, 3 );
      s = dec2str_s( -pa.level, s, 3 ); s = strcopy( s, "dB" );
      
      s = strcopy( s, "\n On" );
      s = putcursor( s, 4 );
      s = dec2str_u( pa.oncycle, s, 2 );
      s = strcopy( s, "   Off" );
      s = putcursor( s, 5 );
      s = dec2str_u( pa.offcycle, s, 2 );
      
    }else{          // Twotone & sin 2ch
      s = putcursor( str, 2 );
      s = putfreq1( s );
      s = putcursor( s, 3 );
      s = uint32str( pa.freq2, s, 5 ); s = strcopy( s, "Hz\n  " );
      
      s = putcursor( s, 4 );
      s = dec2str_s( -pa.level, s, 3 ); s = strcopy( s, "dB " );
      s = putcursor( s, 5 );
      s = dec2str_s( pa.phase, s, 3 ); s = strcopy( s, "deg" );
    }
  }else{
    s = ( step == 0 )? strcopy( str, "\001" ): str; // J[\}[N
    s = strcopy( s, modestr[mode] );
    s = putcursor( s, 1 );
    if( step >= 1 ) {
      if( mode == 0 ) {
        s = strcopy( s, wavestr[pa.wave] );
      }else if( (mode == 1)||(mode == 2) ) {        // memrd or memwr
        s = dec2str_u( mem, s, 2 );
      }else if( mode == 3 ) {                       // freqstep
        s = strcopy( s, freqstepstr[freqstep] );
      }else if( mode == 4 ) {                       // phasestep
        s = dec2str_s( phasesteptbl[phasestep], s, 2 );
        s = strcopy( s, "deg" );
      }else{                                        // levelstep
        s = dec2str_s( levelsteptbl[levelstep], s, 2 );
        s = strcopy( s, "dB" );
      }
    }
    s = strcopy( s, "\n" );
    // Qsڕ\
    if( step>=2 ) {
      switch( pa.wave ) {
        case 0:       // sin
        case 1:       // square
        case 2:       // saw
        case 3:       // triangle
          s = putcursor( s, 2 );       // J[\
          s = uint32str( pa.freq1, s, 5 ); s = strcopy( s, "Hz  " );
          s = putcursor( s, 3 );          // J[\
          s = dec2str_s( -pa.level, s, 3 ); strcopy( s, "dB" );
          break;
        case 7:       // white noise
        case 8:       // pink noise
          s = dec2str_s( -pa.level, s, 14 ); strcopy( s, "dB" );
          break;
      }
    }else if( (step==1)&&((mode==1)||(mode==2)) ) { // mem ̂Ƃ͐ݒf[^\ 
      if( checksetting() ) {
        s = strcopy( s, "No Data" );
      }else{
        s = strcopy( s, waveshortstr[pa.wave] );
        if( (pa.wave!=7)&&(pa.wave!=8) ) s = uint32str( pa.freq1, s, 5 );
        if( (pa.wave==5)||(pa.wave==6) ) s = uint32str( pa.freq2, s, 5 );
        s = dec2str_s( -pa.level, s, 3 );
      }
    }
  }
  lcdc_putstr( str );
}


// ------------------------------------------------------------------------
// C
// ------------------------------------------------------------------------
int main(void) {
  avr_init();
  lcdc_init();
  lcdc_putstr( "AudioSG 20070403\n  ojinsan koubou" );
  lcdc_setcg( 1, "\000\004\006\007\006\004\000\000" );  //EOp
  
  memread( 0 );       // f[^ǂݍ
  
  while(1) {
    if( checksetting() ) defaultsetting();
    sgsetting();
    
    // Mo
    switch( pa.wave ) {
      case 0: sine(); break;            // sine
      case 1:                           // square
      case 2:                           // saw
      case 3: etcwave(); break;         // triangle
      case 4: toneburst(); break;       // tone burst
      case 5: sin2ch(); break;          // sine 2ch
      case 6: twotone(); break;         // two tone
      case 7: whitenoise(); break;      // white noise
      case 8:                           // pink noise
        UBRR0 = 1;            // 1:fclk/4
        pinknoise_test();
        UBRR0 = 0;            // 0:fclk/2
        break;
    }
    keyinput();                         // L[͂ɉ
    disp();                             // LCD Ɍݏ\
    
    while( IS_ANYSW_PRESSED ) {         // L[Ăԑ҂
      dly_usec(10000);
      TEST_H;
      dly_usec(10000);
      TEST_L;
    }
  }
  
  return(0);
}
